@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <string.h>
 #include <net/if.h>
 
 #include <gdbus.h>
@@ -31,7 +32,8 @@
 
 struct gateway_data {
 	int index;
-	char *gateway;
+	char *ipv4_gateway;
+	char *ipv6_gateway;
 	struct connman_element *element;
 	unsigned int order;
 	gboolean active;
@@ -53,11 +55,12 @@ static struct gateway_data *find_gateway(int index, const char *gateway)
 	for (list = gateway_list; list; list = list->next) {
 		struct gateway_data *data = list->data;
 
-		if (data->gateway == NULL)
+		if (data->ipv4_gateway == NULL)
 			continue;
 
 		if (data->index == index &&
-				g_str_equal(data->gateway, gateway) == TRUE)
+				g_str_equal(data->ipv4_gateway, gateway)
+								== TRUE)
 			return data;
 	}
 
@@ -69,15 +72,19 @@ static int del_routes(struct gateway_data *data)
 	if (data->vpn) {
 		if (data->vpn_phy_index >= 0)
 			connman_inet_del_host_route(data->vpn_phy_index,
-							data->gateway);
+							data->ipv4_gateway);
 		return connman_inet_clear_gateway_address(data->index,
 							data->vpn_ip);
-	} else if (g_strcmp0(data->gateway, "0.0.0.0") == 0) {
+	} else if (g_strcmp0(data->ipv4_gateway, "0.0.0.0") == 0) {
 		return connman_inet_clear_gateway_interface(data->index);
 	} else {
-		connman_inet_del_host_route(data->index, data->gateway);
+		connman_inet_del_ipv6_host_route(data->index,
+						data->ipv6_gateway);
+		connman_inet_clear_ipv6_gateway_address(data->index,
+							data->ipv6_gateway);
+		connman_inet_del_host_route(data->index, data->ipv4_gateway);
 		return connman_inet_clear_gateway_address(data->index,
-							data->gateway);
+							data->ipv4_gateway);
 	}
 }
 
@@ -96,17 +103,25 @@ static void find_element(struct connman_element *element, gpointer user_data)
 	data->element = element;
 }
 
-static struct gateway_data *add_gateway(int index, const char *gateway)
+static struct gateway_data *add_gateway(int index, const char *gateway,
+						const char *ipv6_gateway)
 {
 	struct gateway_data *data;
 	struct connman_service *service;
 
+	DBG("index %d ipv4 gateway %s ipv6 gateway %s", index, gateway,
+							ipv6_gateway);
+
+	if (strlen(gateway) == 0)
+		return NULL;
+
 	data = g_try_new0(struct gateway_data, 1);
 	if (data == NULL)
 		return NULL;
 
 	data->index = index;
-	data->gateway = g_strdup(gateway);
+	data->ipv4_gateway = g_strdup(gateway);
+	data->ipv6_gateway = g_strdup(ipv6_gateway);
 	data->active = FALSE;
 	data->element = NULL;
 	data->vpn_ip = NULL;
@@ -142,7 +157,7 @@ static void set_default_gateway(struct gateway_data *data)
 	struct connman_element *element = data->element;
 	struct connman_service *service = NULL;
 
-	DBG("gateway %s", data->gateway);
+	DBG("gateway %s", data->ipv4_gateway);
 
 	if (data->vpn == TRUE) {
 		connman_inet_set_gateway_address(data->index, data->vpn_ip);
@@ -151,15 +166,16 @@ static void set_default_gateway(struct gateway_data *data)
 		return;
 	}
 
-	if (g_strcmp0(data->gateway, "0.0.0.0") == 0) {
+	if (g_strcmp0(data->ipv4_gateway, "0.0.0.0") == 0) {
 		if (connman_inet_set_gateway_interface(element->index) < 0)
 			return;
 		goto done;
 	}
 
-	connman_inet_add_host_route(element->index, data->gateway);
-
-	if (connman_inet_set_gateway_address(element->index, data->gateway) < 0)
+	connman_inet_set_ipv6_gateway_address(element->index,
+						data->ipv6_gateway);
+	if (connman_inet_set_gateway_address(element->index,
+					data->ipv4_gateway) < 0)
 		return;
 
 done:
@@ -189,7 +205,7 @@ static int remove_gateway(struct gateway_data *data)
 {
 	int err;
 
-	DBG("gateway %s", data->gateway);
+	DBG("gateway %s", data->ipv4_gateway);
 
 	gateway_list = g_slist_remove(gateway_list, data);
 
@@ -198,7 +214,7 @@ static int remove_gateway(struct gateway_data *data)
 	else
 		err = 0;
 
-	g_free(data->gateway);
+	g_free(data->ipv4_gateway);
 	g_free(data->vpn_ip);
 	g_free(data);
 
@@ -245,8 +261,9 @@ static struct gateway_data *find_active_gateway(void)
 static int connection_probe(struct connman_element *element)
 {
 	struct connman_service *service = NULL;
-	const char *gateway = NULL;
+	const char *gateway = NULL, *ipv6_gateway = NULL;
 	const char *vpn_ip = NULL;
+	const char *domainname = NULL;
 	struct gateway_data *active_gateway = NULL;
 	struct gateway_data *new_gateway = NULL;
 
@@ -261,11 +278,17 @@ static int connection_probe(struct connman_element *element)
 
 	connman_element_get_value(element,
 				CONNMAN_PROPERTY_ID_IPV4_GATEWAY, &gateway);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV6_GATEWAY, &ipv6_gateway);
 
 	connman_element_get_value(element,
-				  CONNMAN_PROPERTY_ID_IPV4_ADDRESS, &vpn_ip);
+			CONNMAN_PROPERTY_ID_IPV4_ADDRESS, &vpn_ip);
 
-	DBG("gateway %s", gateway);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_DOMAINNAME, &domainname);
+
+	DBG("ipv4 gateway %s ipv6 gateway %s domainname %s",
+				gateway, ipv6_gateway, domainname);
 
 	/*
 	 * If gateway is NULL, it's a point to point link and the default
@@ -276,14 +299,24 @@ static int connection_probe(struct connman_element *element)
 		element->ipv4.gateway = g_strdup(gateway);
 	}
 
-	service = __connman_element_get_service(element);
-	__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_READY);
-
 	connman_element_set_enabled(element, TRUE);
 
 	active_gateway = find_active_gateway();
-	new_gateway = add_gateway(element->index, gateway);
+	new_gateway = add_gateway(element->index, gateway, ipv6_gateway);
+	if (new_gateway == NULL)
+		return 0;
+
+	service = __connman_element_get_service(element);
+
+	connman_inet_add_ipv6_host_route(element->index,
+					new_gateway->ipv6_gateway, NULL);
+	connman_inet_add_host_route(element->index,
+					new_gateway->ipv4_gateway, NULL);
+	__connman_service_nameserver_add_routes(service,
+						new_gateway->ipv4_gateway);
+	__connman_service_set_domainname(service, domainname);
+
+	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY);
 
 	if (service == NULL) {
 		new_gateway->vpn = TRUE;
@@ -301,9 +334,9 @@ static int connection_probe(struct connman_element *element)
 	}
 
 	if (new_gateway->vpn == TRUE) {
-		connman_inet_add_host_route_vpn(active_gateway->index,
-						active_gateway->gateway,
-						new_gateway->gateway);
+		connman_inet_add_host_route(active_gateway->index,
+						new_gateway->ipv4_gateway,
+						active_gateway->ipv4_gateway);
 	}
 
 	if (new_gateway->order >= active_gateway->order) {
@@ -325,6 +358,7 @@ static void connection_remove(struct connman_element *element)
 	DBG("element %p name %s", element, element->name);
 
 	service = __connman_element_get_service(element);
+	__connman_service_nameserver_del_routes(service);
 	__connman_service_indicate_state(service,
 					CONNMAN_SERVICE_STATE_DISCONNECT);
 
@@ -345,8 +379,8 @@ static void connection_remove(struct connman_element *element)
 	set_default = data->vpn;
 
 	if (data->vpn == TRUE && data->vpn_phy_index >= 0)
-		connman_inet_del_host_route(data->vpn_phy_index, data->gateway);
-
+		connman_inet_del_host_route(data->vpn_phy_index,
+						data->ipv4_gateway);
 	err = remove_gateway(data);
 
 	/* with vpn this will be called after the network was deleted,
@@ -392,9 +426,9 @@ void __connman_connection_cleanup(void)
 	for (list = gateway_list; list; list = list->next) {
 		struct gateway_data *data = list->data;
 
-		DBG("index %d gateway %s", data->index, data->gateway);
+		DBG("index %d gateway %s", data->index, data->ipv4_gateway);
 
-		g_free(data->gateway);
+		g_free(data->ipv4_gateway);
 		g_free(data);
 		list->data = NULL;
 	}
