@@ -194,53 +194,80 @@ static void powered_changed(struct connman_device *device)
 					DBUS_TYPE_BOOLEAN, &device->powered);
 }
 
-static int set_powered(struct connman_device *device, connman_bool_t powered)
+int __connman_device_enable(struct connman_device *device)
 {
-	struct connman_device_driver *driver = device->driver;
 	int err;
 
-	DBG("device %p powered %d", device, powered);
+	DBG("device %p", device);
 
-	if (device->powered_pending == powered)
+	if (!device->driver || !device->driver->enable)
+		return -EOPNOTSUPP;
+
+	if (device->powered_pending == TRUE)
 		return -EALREADY;
 
-	if (!driver)
-		return -EINVAL;
+	if (device->blocked == TRUE)
+		return -ENOLINK;
 
-	if (powered == TRUE) {
-		if (driver->enable) {
-			device->powered_pending = powered;
-
-			err = driver->enable(device);
-			if (err == 0)
-				__connman_technology_enable_device(device);
-		} else
-			err = -EINVAL;
-	} else {
-		device->powered_pending = powered;
-
-		device->reconnect = FALSE;
-
-		clear_scan_trigger(device);
-
-		g_hash_table_remove_all(device->networks);
-
-		if (driver->disable) {
-			err = driver->disable(device);
-			if (err == 0)
-				__connman_technology_disable_device(device);
-		} else
-			err = -EINVAL;
+	err = device->driver->enable(device);
+	if (err < 0) {
+		if (err == -EINPROGRESS)
+			device->powered_pending = TRUE;
+		return err;
 	}
 
-	if (err == 0) {
-		device->powered = powered;
+	device->powered_pending = TRUE;
+	device->powered = TRUE;
 
-		if (device->registered == TRUE)
-			powered_changed(device);
+	__connman_technology_enable_device(device);
+
+	return 0;
+}
+
+int __connman_device_disable(struct connman_device *device)
+{
+	int err;
+
+	DBG("device %p", device);
+
+	if (!device->driver || !device->driver->disable)
+		return -EOPNOTSUPP;
+
+	if (device->powered == FALSE)
+		return -ENOLINK;
+
+	if (device->powered_pending == FALSE)
+		return -EALREADY;
+
+	device->reconnect = FALSE;
+
+	clear_scan_trigger(device);
+
+	g_hash_table_remove_all(device->networks);
+
+	err = device->driver->disable(device);
+	if (err < 0) {
+		if (err == -EINPROGRESS)
+			device->powered_pending = FALSE;
+		return err;
 	}
 
-	return err;
+	device->powered_pending = FALSE;
+	device->powered = FALSE;
+
+	__connman_technology_disable_device(device);
+
+	return 0;
+}
+
+static int set_powered(struct connman_device *device, connman_bool_t powered)
+{
+	DBG("device %p powered %d", device, powered);
+
+	if (powered == TRUE)
+		return __connman_device_enable(device);
+	else
+		return __connman_device_disable(device);
 }
 
 void __connman_device_list(DBusMessageIter *iter, void *user_data)
@@ -656,8 +683,6 @@ static void unregister_network(gpointer data)
 	connman_element_unregister((struct connman_element *) network);
 
 	connman_network_unref(network);
-
-	__connman_network_set_device(network, NULL);
 }
 
 static void device_destruct(struct connman_element *element)
@@ -1031,74 +1056,24 @@ int __connman_device_scan(struct connman_device *device)
 	return device->driver->scan(device);
 }
 
-int __connman_device_enable(struct connman_device *device)
+int __connman_device_enable_persistent(struct connman_device *device)
 {
 	int err;
 
 	DBG("device %p", device);
 
-	if (!device->driver || !device->driver->enable)
-		return -EOPNOTSUPP;
-
-	if (device->powered_pending == TRUE)
-		return -EALREADY;
-
-	device->powered_pending = TRUE;
-
-	err = device->driver->enable(device);
-	if (err < 0)
-		return err;
-
-	device->powered = TRUE;
-
-	__connman_technology_enable_device(device);
-
-	return 0;
-}
-
-int __connman_device_enable_persistent(struct connman_device *device)
-{
-	DBG("device %p", device);
-
 	device->powered_persistent = TRUE;
 
 	__connman_storage_save_device(device);
 
-	return __connman_device_enable(device);
-}
+	err = __connman_device_enable(device);
+	if (err == 0 || err == -EINPROGRESS) {
+		if (__connman_profile_get_offlinemode() == TRUE)
+			__connman_profile_set_offlinemode(FALSE, FALSE);
 
-int __connman_device_disable(struct connman_device *device)
-{
-	int err;
+	}
 
-	DBG("device %p", device);
-
-	if (!device->driver || !device->driver->disable)
-		return -EOPNOTSUPP;
-
-	if (device->powered == FALSE)
-		return -ENOLINK;
-
-	if (device->powered_pending == FALSE)
-		return -EALREADY;
-
-	device->powered_pending = FALSE;
-
-	device->reconnect = FALSE;
-
-	clear_scan_trigger(device);
-
-	g_hash_table_remove_all(device->networks);
-
-	err = device->driver->disable(device);
-	if (err < 0)
-		return err;
-
-	device->powered = FALSE;
-
-	__connman_technology_disable_device(device);
-
-	return 0;
+	return err;
 }
 
 int __connman_device_disable_persistent(struct connman_device *device)
@@ -1186,24 +1161,6 @@ static void scanning_changed(struct connman_device *device)
 					DBUS_TYPE_BOOLEAN, &device->scanning);
 }
 
-static void mark_network_available(gpointer key, gpointer value,
-						gpointer user_data)
-{
-	struct connman_network *network = value;
-
-	connman_network_set_available(network, TRUE);
-}
-
-void connman_device_cleanup_scanning(struct connman_device *device)
-{
-	device->scanning = FALSE;
-
-	scanning_changed(device);
-
-	g_hash_table_foreach(device->networks,
-				mark_network_available, NULL);
-}
-
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1468,6 +1425,9 @@ void __connman_device_set_network(struct connman_device *device,
 {
 	const char *name;
 
+	if (device == NULL)
+		return;
+
 	if (device->network == network)
 		return;
 
